/*
 * Python bindings module for libregf (pyregf)
 *
 * Copyright (c) 2009-2014, Joachim Metz <joachim.metz@gmail.com>
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <types.h>

#if defined( HAVE_STDLIB_H ) || defined( HAVE_WINAPI )
#include <stdlib.h>
#endif

#include "pyregf.h"
#include "pyregf_error.h"
#include "pyregf_file.h"
#include "pyregf_file_object_io_handle.h"
#include "pyregf_key.h"
#include "pyregf_keys.h"
#include "pyregf_libcerror.h"
#include "pyregf_libcstring.h"
#include "pyregf_libregf.h"
#include "pyregf_python.h"
#include "pyregf_unused.h"
#include "pyregf_value.h"
#include "pyregf_value_types.h"
#include "pyregf_values.h"

#if !defined( LIBREGF_HAVE_BFIO )
LIBREGF_EXTERN \
int libregf_check_file_signature_file_io_handle(
     libbfio_handle_t *file_io_handle,
     libregf_error_t **error );
#endif

/* The pyregf module methods
 */
PyMethodDef pyregf_module_methods[] = {
	{ "get_version",
	  (PyCFunction) pyregf_get_version,
	  METH_NOARGS,
	  "get_version() -> String\n"
	  "\n"
	  "Retrieves the version." },

	{ "check_file_signature",
	  (PyCFunction) pyregf_check_file_signature,
	  METH_VARARGS | METH_KEYWORDS,
	  "check_file_signature(filename) -> Boolean\n"
	  "\n"
	  "Checks if a file has a Windows NT Registry File (REGF) signature." },

	{ "check_file_signature_file_object",
	  (PyCFunction) pyregf_check_file_signature_file_object,
	  METH_VARARGS | METH_KEYWORDS,
	  "check_file_signature_file_object(file_object) -> Boolean\n"
	  "\n"
	  "Checks if a file has a Windows NT Registry File (REGF) signature using a file-like object." },

	{ "open",
	  (PyCFunction) pyregf_file_new_open,
	  METH_VARARGS | METH_KEYWORDS,
	  "open(filename, mode='r') -> Object\n"
	  "\n"
	  "Opens a file." },

	{ "open_file_object",
	  (PyCFunction) pyregf_file_new_open_file_object,
	  METH_VARARGS | METH_KEYWORDS,
	  "open_file_object(file_boject, mode='r') -> Object\n"
	  "\n"
	  "Opens a file using a file-like object." },

	/* Sentinel */
	{ NULL,
	  NULL,
	  0,
	  NULL}
};

/* Retrieves the pyregf/libregf version
 * Returns a Python object if successful or NULL on error
 */
PyObject *pyregf_get_version(
           PyObject *self PYREGF_ATTRIBUTE_UNUSED,
           PyObject *arguments PYREGF_ATTRIBUTE_UNUSED )
{
	const char *errors           = NULL;
	const char *version_string   = NULL;
	size_t version_string_length = 0;

	PYREGF_UNREFERENCED_PARAMETER( self )
	PYREGF_UNREFERENCED_PARAMETER( arguments )

	Py_BEGIN_ALLOW_THREADS

	version_string = libregf_get_version();

	Py_END_ALLOW_THREADS

	version_string_length = libcstring_narrow_string_length(
	                         version_string );

	/* Pass the string length to PyUnicode_DecodeUTF8
	 * otherwise it makes the end of string character is part
	 * of the string
	 */
	return( PyUnicode_DecodeUTF8(
	         version_string,
	         (Py_ssize_t) version_string_length,
	         errors ) );
}

/* Checks if the file has a Windows NT Registry File (REGF) signature
 * Returns a Python object if successful or NULL on error
 */
PyObject *pyregf_check_file_signature(
           PyObject *self PYREGF_ATTRIBUTE_UNUSED,
           PyObject *arguments,
           PyObject *keywords )
{
	libcerror_error_t *error    = NULL;
	static char *function       = "pyregf_check_file_signature";
	static char *keyword_list[] = { "filename", NULL };
	const char *filename        = NULL;
	int result                  = 0;

	PYREGF_UNREFERENCED_PARAMETER( self )

	if( PyArg_ParseTupleAndKeywords(
	     arguments,
	     keywords,
	     "|s",
	     keyword_list,
	     &filename ) == 0 )
	{
		return( NULL );
	}
	Py_BEGIN_ALLOW_THREADS

	result = libregf_check_file_signature(
	          filename,
	          &error );

	Py_END_ALLOW_THREADS

	if( result == -1 )
	{
		pyregf_error_raise(
		 error,
		 PyExc_IOError,
		 "%s: unable to check file signature.",
		 function );

		libcerror_error_free(
		 &error );

		return( NULL );
	}
	if( result != 0 )
	{
		return( Py_True );
	}
	return( Py_False );
}

/* Checks if the file has a Windows NT Registry File (REGF) signature using a file-like object
 * Returns a Python object if successful or NULL on error
 */
PyObject *pyregf_check_file_signature_file_object(
           PyObject *self PYREGF_ATTRIBUTE_UNUSED,
           PyObject *arguments,
           PyObject *keywords )
{
	libcerror_error_t *error         = NULL;
	libbfio_handle_t *file_io_handle = NULL;
	PyObject *file_object            = NULL;
	static char *function            = "pyregf_check_file_signature_file_object";
	static char *keyword_list[]      = { "file_object", NULL };
	int result                       = 0;

	PYREGF_UNREFERENCED_PARAMETER( self )

	if( PyArg_ParseTupleAndKeywords(
	     arguments,
	     keywords,
	     "|O",
	     keyword_list,
	     &file_object ) == 0 )
	{
		return( NULL );
	}
	if( pyregf_file_object_initialize(
	     &file_io_handle,
	     file_object,
	     &error ) != 1 )
	{
		pyregf_error_raise(
		 error,
		 PyExc_MemoryError,
		 "%s: unable to initialize file IO handle.",
		 function );

		libcerror_error_free(
		 &error );

		goto on_error;
	}
	Py_BEGIN_ALLOW_THREADS

	result = libregf_check_file_signature_file_io_handle(
	          file_io_handle,
	          &error );

	Py_END_ALLOW_THREADS

	if( result == -1 )
	{
		pyregf_error_raise(
		 error,
		 PyExc_IOError,
		 "%s: unable to check file signature.",
		 function );

		libcerror_error_free(
		 &error );

		goto on_error;
	}
	if( libbfio_handle_free(
	     &file_io_handle,
	     &error ) != 1 )
	{
		pyregf_error_raise(
		 error,
		 PyExc_MemoryError,
		 "%s: unable to free file IO handle.",
		 function );

		libcerror_error_free(
		 &error );

		goto on_error;
	}
	if( result != 0 )
	{
		return( Py_True );
	}
	return( Py_False );

on_error:
	if( file_io_handle != NULL )
	{
		libbfio_handle_free(
		 &file_io_handle,
		 NULL );
	}
	return( NULL );
}

/* Declarations for DLL import/export
 */
#ifndef PyMODINIT_FUNC
#define PyMODINIT_FUNC void
#endif

/* Initializes the pyregf module
 */
PyMODINIT_FUNC initpyregf(
                void )
{
	PyObject *module                      = NULL;
	PyTypeObject *file_type_object        = NULL;
	PyTypeObject *key_type_object         = NULL;
	PyTypeObject *keys_type_object        = NULL;
	PyTypeObject *value_type_object       = NULL;
	PyTypeObject *value_types_type_object = NULL;
	PyTypeObject *values_type_object      = NULL;
	PyGILState_STATE gil_state            = 0;

/*
	libregf_notify_set_stream(
	 stderr,
	 NULL );
	libregf_notify_set_verbose(
	 1 );
*/

	/* Create the module
	 * This function must be called before grabbing the GIL
	 * otherwise the module will segfault on a version mismatch
	 */
	module = Py_InitModule3(
	          "pyregf",
	          pyregf_module_methods,
	          "Python libregf module (pyregf)." );

	PyEval_InitThreads();

	gil_state = PyGILState_Ensure();

	/* Setup the file type object
	 */
	pyregf_file_type_object.tp_new = PyType_GenericNew;

	if( PyType_Ready(
	     &pyregf_file_type_object ) < 0 )
	{
		goto on_error;
	}
	Py_IncRef(
	 (PyObject *) &pyregf_file_type_object );

	file_type_object = &pyregf_file_type_object;

	PyModule_AddObject(
	 module,
	 "file",
	 (PyObject *) file_type_object );

	/* Setup the keys type object
	 */
	pyregf_keys_type_object.tp_new = PyType_GenericNew;

	if( PyType_Ready(
	     &pyregf_keys_type_object ) < 0 )
	{
		goto on_error;
	}
	Py_IncRef(
	 (PyObject *) &pyregf_keys_type_object );

	keys_type_object = &pyregf_keys_type_object;

	PyModule_AddObject(
	 module,
	 "_keys",
	 (PyObject *) keys_type_object );

	/* Setup the key type object
	 */
	pyregf_key_type_object.tp_new = PyType_GenericNew;

	if( PyType_Ready(
	     &pyregf_key_type_object ) < 0 )
	{
		goto on_error;
	}
	Py_IncRef(
	 (PyObject *) &pyregf_key_type_object );

	key_type_object = &pyregf_key_type_object;

	PyModule_AddObject(
	 module,
	 "key",
	 (PyObject *) key_type_object );

	/* Setup the values type object
	 */
	pyregf_values_type_object.tp_new = PyType_GenericNew;

	if( PyType_Ready(
	     &pyregf_values_type_object ) < 0 )
	{
		goto on_error;
	}
	Py_IncRef(
	 (PyObject *) &pyregf_values_type_object );

	values_type_object = &pyregf_values_type_object;

	PyModule_AddObject(
	 module,
	 "_values",
	 (PyObject *) values_type_object );

	/* Setup the value type object
	 */
	pyregf_value_type_object.tp_new = PyType_GenericNew;

	if( PyType_Ready(
	     &pyregf_value_type_object ) < 0 )
	{
		goto on_error;
	}
	Py_IncRef(
	 (PyObject *) &pyregf_value_type_object );

	value_type_object = &pyregf_value_type_object;

	PyModule_AddObject(
	 module,
	 "value",
	 (PyObject *) value_type_object );

	/* Setup the value types type object
	 */
	pyregf_value_types_type_object.tp_new = PyType_GenericNew;

	if( pyregf_value_types_init_type(
             &pyregf_value_types_type_object ) != 1 )
	{
		goto on_error;
	}
	if( PyType_Ready(
	     &pyregf_value_types_type_object ) < 0 )
	{
		goto on_error;
	}
	Py_IncRef(
	 (PyObject *) &pyregf_value_types_type_object );

	value_types_type_object = &pyregf_value_types_type_object;

	PyModule_AddObject(
	 module,
	 "value_types",
	 (PyObject *) value_types_type_object );

on_error:
	PyGILState_Release(
	 gil_state );
}

